#Python uses references, actual pointers are almost never needed.
#Use this library only if you have a very special use case (like injecting python interpreter into a process and manipulating the process memory)


import ctypes
import sys


mem_backups = []
backup_counter = 0


class Pointer():          
    def _refer_to_cdata(self, pointable, implicit_object = True):
        self.garbage_lock1 = pointable # protect from garbage collection
        self.pointer_type = type(pointable)
        self._address = ctypes.addressof(pointable)
        self.c_memory = True
        self.implicit_object = implicit_object
        
    def refer_to_cdata(self, cdata):
        self._refer_to_cdata(cdata, implicit_object = False)
        
    def refer_to(self, obj):
        # py_object internally stores a second level pointer to obj
        # Address of the first level pointer can be obtained with ctypes.addressof(py_object)
        # Then, this pointer can be used to get back obj
        py_object = ctypes.py_object(obj)
        self._refer_to_cdata(py_object)
        
    def change_address(self, address, ctype = None):
        self._address = address
        if ctype is None:
            c_memory = False
        else:
            c_memory = True
        self.c_memory = c_memory
        self.implicit_object = False
        self.pointer_type = ctype

    def get_address(self):
        if  self.c_memory == False:
            address = ctypes.cast(self._address, ctypes.c_void_p).value
        else:
            if self.implicit_object:
                first_level = ctypes.cast(self._address, ctypes.POINTER(ctypes.c_void_p)).contents
                address = first_level.value
            else:
                address = self._address
        return address

    def dereference(self):
        if self.c_memory == False:           
            return ctypes.cast(self._address, ctypes.py_object).value
        else:
            pointer = ctypes.cast(self._address, ctypes.POINTER(self.pointer_type))
            c_data = pointer.contents
            if self.implicit_object == True:
                return c_data.value
            return c_data

    def get_size(self):
        value = self.dereference()
        if self.c_memory and not self.implicit_object:
            size = ctypes.sizeof(value)
        else:
            size = sys.getsizeof(value)
        return size

    def get_mem(self):
        size = self.get_size()
        memory = mem_read(self.get_address(), size)
        return memory

    def overwrite_mem(self, data, force_write = False):
        """Passing True to force_write may not do any observable harm if the overflow is small but this is a very bad habit. It's undefined behavior after all."""
        size_limit = self.get_size()
        if len(data) > size_limit and not force_write:
            error = "Size of data argument "
            error = error + "(" + str(len(data)) + ") "
            error = error + "is bigger than size of the pointed object "
            error = error + "(" + str(size_limit) + ")\n"
            error = error + "If you are sure of allocated memory is enough, pass force_write = True.\n"
            error = error + "Misuse of this argument can cause a buffer overflow and crash the python interpreter"
            raise MemoryError(error)
        backup_id = mem_write(self.get_address(), data = data)
        return backup_id
    
    def change_value(self, value, force_write = False):
        """Byte copy of value to the pointed address

        This is a very powerful function. But great power brings great responsibility.
        If you modify immutable literals (yes you can do it), you will never be able to use them again but their modified forms.
        This can mess up not only your main script, but also other modules you use. Literals are shared through the whole runtime.
        Only way to solve this problem is calling mem_restore() function."""
        if type(value) != type(self.dereference()) and self.c_memory and not self.implicit_object:
            raise TypeError("type of value argument is different than type of the ctypes object which hold the address of the C object this Pointer instance refers to.")
        self.garbage_lock2 = value
        if self.c_memory and not self.implicit_object:
            pointer = pointer_to_object(value, c_object = True)
        else:
            pointer = pointer_to_object(value)
        self.garbage_lock3 = pointer
        new_mem = pointer.get_mem()
        backup_id = self.overwrite_mem(new_mem, force_write = force_write)        
        return backup_id

    class ChangeValue:
        def __init__(self, owner, value, force_write = False):            
            self.owner = owner
            self.value = value
            self.force_write = force_write
            
        def __enter__(self):
            self.backup_id = self.owner.change_value(self.value, force_write = self.force_write)
            return self.owner
        
        def __exit__(self, exc_type, exc_value, traceback):
            mem_restore(self.backup_id)

    def temp_value(self, value, force_write = False):
        """Return a context manager for change_value() method. Memory will always be restored on exit."""
        temp_changer = self.ChangeValue(self, value, force_write = force_write)
        return temp_changer
            
    @property
    def value(self):
        return self.dereference()

    @value.setter
    def value(self, new_value):
        return self.change_value(new_value)


def pointer_to_object(obj, c_object = False):
    pointer = Pointer()
    if c_object == False:
        pointer.refer_to(obj)
    else:
        try:
            pointer.refer_to_cdata(obj)
        except TypeError:
            raise TypeError("type of obj must be a ctypes type when c_object is True")
    return pointer


def pointer_to_address(address, ctype = None):
    pointer = Pointer()
    pointer.change_address(address, ctype = ctype)
    return pointer


def dereference(pointer, *value, force_write = False):
    """This function acts like dereference operator (*) of C"""
    if len(value) > 1:
        raise TypeError(f"dereference() takes max 2 positional arguments but {len(value)+1} was given")
    if not value:
        return pointer.dereference()
    else:
        return pointer.change_value(value[0], force_write = force_write)


def address_of(obj, c_object = False):
    """This function acts like address-of operator (&) of C"""
    pointer = pointer_to_object(obj, c_object = c_object)
    return pointer.get_address()


def get_memory_proxy(address, array_type = ctypes.c_char * 1):
    pointer = pointer_to_address(address, ctype = array_type)
    data = dereference(pointer)
    return data


def mem_read(address, amount = 1):
    data = get_memory_proxy(address, array_type = ctypes.c_char * amount)
    return data[:]


def mem_write(address, data = b""):
    current_data = get_memory_proxy(address, array_type = ctypes.c_ubyte * len(data))
    global mem_backups
    global backup_counter
    backup = list(current_data)
    backup_id = backup_counter
    backup_counter += 1
    mem_backups.append([current_data, backup])
    current_data[:] = data
    return backup_id

    
def mem_restore(backup_id):
    global mem_backups
    global backup_counter    
    info = mem_backups[backup_id]
    del mem_backups[backup_id]
    backup_counter -= 1 
    backup = info.pop()
    memory_region = info.pop()
    memory_region[:] = backup


def mem_restore_last():
    global mem_backups
    global backup_counter
    info = mem_backups.pop()
    backup_counter -= 1
    backup = info.pop()
    memory_region = info.pop()
    memory_region[:] = backup    
    
    




from pyptrs import *
import ctypes


def test1():
    x = 1234
    pointer = pointer_to_object(x)
    if dereference(pointer) is x:        
        return True
    else:
        return False


def test2():
    x = [1,2,3]
    address = address_of(x)
    pointer = pointer_to_address(address)
    dereference(pointer).append(4)
    if dereference(pointer) is x:
        return  True
    else:
        return False


def test3():
    """Equivalent c++ code:

    int pointer_test2()
    {
        int x[4] = { 1,2,3,0 };
        int(*pointer)[4] = &x;
        (*pointer)[3] = 4;
        for (int i = 0; i < 4; i++)
        {
            if ((*pointer)[i] != x[i])
            {
                return false;
            }
        }
        return true;
    }
    
    """
    x = [1,2,3,0]
    address = address_of(x)
    pointer = pointer_to_address(address)
    dereference(pointer)[3] = 4
    for i in range(4):
        if not (dereference(pointer)[i] is x[i]):            
            return False
    return True


def test4():
    a = 24532
    p = pointer_to_object(a)
    c_memory = p._address
    p2 = pointer_to_address(c_memory, ctype = ctypes.py_object)
    if p2.dereference().value == a:
        return True
    else:
        return False


def test5():
    x = [1,2,3]
    pointer = pointer_to_address(address_of(x))
    if dereference(pointer) is x:
        return True
    else:
        return False


def test6():
    a = 0xaaaaaaaaaaaaa
    pointer = pointer_to_object(a)
    b = 0xfffffffffffff
    dereference(pointer, b)
    if a == b:
        mem_restore_last()
        return True
    else:
        return False

def test7():
    a = "abc"
    pointer = pointer_to_object(a)
    pointer.change_value("def")
    if a == "def":
        mem_restore_last()
        return True
    else:
        return False


def test8():
    a = bytes(b"abc")
    pointer = pointer_to_object(a)
    pointer.change_value(bytes(b"def"))
    if a == bytes(b"def"):
        mem_restore_last()
        return True
    else:
        return False


def test9():
    a = bytes(b"abc")
    pointer = pointer_to_object(a)
    pointer.change_value(12354087532633467)
    if a == 12354087532633467:
        mem_restore_last()
        return True
    else:
        return False


def test10():    
    a = 0x1235ab
    b = ctypes.c_int(a)
    pointer = pointer_to_object(b, c_object = True)
    pointer.value = ctypes.c_int(5)
    if b.value == 5:
        return True
    else:
        return False


def test11():
    pointer = pointer_to_object("a" * 4096) 
    pointer.change_value("b")
    pointer2 = pointer_to_object("c" * 4097) 
    pointer2.change_value("d")
    var = "e" * 4097
    pointer3 = pointer_to_object(var) 
    pointer3.change_value("f")
    if "a" * 4096 == "b" and "c" * 4097 != "d" and var == "f":
        mem_restore_last()
        mem_restore_last()
        mem_restore_last()
        return True
    else:
        return False


def test12():    
    pointer = pointer_to_object("abc")
    backup_id = pointer.change_value("def")
    if "abc" == "def":
        mem_restore(backup_id)
        return True
    else:
        return False

def test13():
    pointer = pointer_to_object(True)
    backup_id = pointer.change_value(False)
    if True == False:
        mem_restore(backup_id)
        return True
    else:
        return False


def test14():
    a = [1,2,3]
    pointer = pointer_to_object(a)
    backup_id = pointer.change_value([4,5,6])
    if a == [4,5,6]:
        mem_restore(backup_id)
        return True
    else:
        return False


def test15(): # doesnt work on some python versions
    class MyClass:
        def __init__(self, var):
            self.var = var    
    a = "a" * 10000
    pointer = pointer_to_object(a)
    x = MyClass(5)
    pointer.change_value(x)
    if a.var == 5: # a != x because ctypes doesn't have original object return but all their attributes are common        
        return True
    else:        
        return False


def test16():    
    a = "abc"
    pointer = pointer_to_object(a)
    pointer.temp_value("def")
    check1 = a == "abc"
    with pointer.temp_value("def") as p:        
        check2 = a == "def"
        check3 = p is pointer
    check4 = a == "abc"
    if check1 and check2 and check3 and check4:
        return True
    else:
        return False


def test17():    
    class MyClass:
        __slots__ = ["var"]
        def __init__(self, var):
            self.var = var
    a = MyClass(10)
    pointer = pointer_to_object(a)
    x = MyClass(20)
    pointer.change_value(x)
    if a.var == 20:        
        return True
    else:
        return False


def test18():
    a = ctypes.c_int(10)
    p1 = pointer_to_address(address_of(a, c_object = True), ctype = ctypes.c_int)
    if p1.dereference().value == 10:
        return True
    else:
        return False
    
    
if __name__ == "__main__":
    if test1() and test2() and test3() and test4() and test5() and test6() and test7() and test8() and test9() and test10() and test11() and test12() and test13() and test14() and test16() and test17() and test18():
        print("[+] all tests are successful")
    else:
        print("some tests are failed")



from .pyptrs import pointer_to_object, pointer_to_address, \
                  dereference, address_of, \
                  mem_read, mem_write, \
                  mem_restore, mem_restore_last
                  
